import re

from autograder_platform.StudentSubmission.ITransformer import ITransformer

class MagicCommandTransformer(ITransformer):
    MAGIC_COMMAND_PREFIXES: re.Pattern = re.compile(r"^[!%].*$")

    def transform(self, string: str) -> str:
        transformed = []

        for line in string.splitlines():
            if self.MAGIC_COMMAND_PREFIXES.match(line.strip()):
                continue

            transformed.append(line)

        return "\n".join(transformed)


class MatplotLibFigTransformer(ITransformer):
    FIG_SHOW_PATTERN: re.Pattern = re.compile(r"^(matplotlib\.pyplot\.show\(\)|plt\.show\(\))$")

    def transform(self, string: str) -> str:
        curFigNumber = 1
        transformed = []
        for line in string.splitlines():
            if self.FIG_SHOW_PATTERN.match(line.strip()):
                line = line.replace("show()", f"savefig('fig_{curFigNumber}.png')")
                curFigNumber += 1

            transformed.append(line)

        return "\n".join(transformed)



import os.path
import shutil
import unittest
from io import StringIO
from unittest.mock import patch

from autograder_platform.StudentSubmission.common import ValidationError
from language_binds.IPython.IPythonSubmission import IPythonSubmission
from .NotebookBuilder import NotebookBuilder


class TestIPythonSubmission(unittest.TestCase):
    TEST_FILE_DIRECTORY: str = "./sandbox"
    VALID_TESTABLE_CELL: str = \
        """
print("VALID_TESTABLE_CELL")
    """

    VALID_CELL: str = \
        """
print("VALID_CELL")
    """

    def setUp(self):
        if os.path.exists(self.TEST_FILE_DIRECTORY):
            shutil.rmtree(self.TEST_FILE_DIRECTORY)

        os.mkdir(self.TEST_FILE_DIRECTORY)

    def tearDown(self):
        if os.path.exists(self.TEST_FILE_DIRECTORY):
            shutil.rmtree(self.TEST_FILE_DIRECTORY)

    def testDiscoverSingleFile(self):
        filename = NotebookBuilder("notebook.ipynb", self.TEST_FILE_DIRECTORY) \
            .addMarkdownCell("# Markdown Cell") \
            .addCodeCell("testableCell", self.VALID_TESTABLE_CELL) \
            .addMarkdownCell("# Markdown Cell") \
            .addCodeCell("validCell", self.VALID_CELL, runnable=False) \
            .addMarkdownCell("# Markdown Cell") \
            .toFile()

        submission = IPythonSubmission()\
            .setSubmissionRoot(self.TEST_FILE_DIRECTORY)\
            .load()

        discoveredFiles = submission.getDiscoveredFiles()

        self.assertIn(filename, discoveredFiles)

        cells = submission.getCells()

        self.assertEqual(2, len(cells))

    def testDiscoverSingleFileNested(self):
        fullpath = os.path.join(self.TEST_FILE_DIRECTORY, "nested")
        os.mkdir(fullpath)
        filename = NotebookBuilder("notebook.ipynb", fullpath) \
            .addMarkdownCell("# Markdown Cell") \
            .addCodeCell("testableCell", self.VALID_TESTABLE_CELL) \
            .addMarkdownCell("# Markdown Cell") \
            .addCodeCell("validCell", self.VALID_CELL, runnable=False) \
            .addMarkdownCell("# Markdown Cell") \
            .toFile()

        submission = IPythonSubmission() \
            .setSubmissionRoot(self.TEST_FILE_DIRECTORY) \
            .load()

        discoveredFiles = submission.getDiscoveredFiles()

        self.assertIn(filename, discoveredFiles)

        cells = submission.getCells()

        self.assertEqual(2, len(cells))

    def testDiscoverManyFiles(self):
        NotebookBuilder("notebook1.ipynb", self.TEST_FILE_DIRECTORY) \
            .addMarkdownCell("# Markdown Cell") \
            .addCodeCell("testableCell", self.VALID_TESTABLE_CELL) \
            .addMarkdownCell("# Markdown Cell") \
            .toFile()

        NotebookBuilder("notebook2.ipynb", self.TEST_FILE_DIRECTORY) \
            .addMarkdownCell("# Markdown Cell") \
            .addCodeCell("testableCell", self.VALID_TESTABLE_CELL) \
            .addMarkdownCell("# Markdown Cell") \
            .toFile()

        with self.assertRaises(ValidationError) as ex:
            IPythonSubmission()\
                .setSubmissionRoot(self.TEST_FILE_DIRECTORY)\
                .load()

        msg = str(ex.exception)

        self.assertIn("expected one `.ipynb` file", msg.lower())


    def testDiscoverNoNotebooks(self):
        with open(os.path.join(self.TEST_FILE_DIRECTORY, "data.txt"), "w") as w:
            w.write("content")

        with self.assertRaises(ValidationError) as ex:
            IPythonSubmission() \
                .setSubmissionRoot(self.TEST_FILE_DIRECTORY) \
                .load()

        msg = str(ex.exception)

        self.assertIn("no `.ipynb` files were submitted!", msg.lower())

    def testNoTestableCells(self):
        NotebookBuilder("notebook.ipynb", self.TEST_FILE_DIRECTORY) \
            .addMarkdownCell("# Markdown Cell") \
            .addCodeCell("validCell", self.VALID_CELL, runnable=False) \
            .addMarkdownCell("# Markdown Cell") \
            .toFile()

        with self.assertRaises(ValidationError) as ex:
            IPythonSubmission() \
                .setSubmissionRoot(self.TEST_FILE_DIRECTORY) \
                .load()

        msg = str(ex.exception)

        self.assertIn("at least one cell to be testable", msg.lower())

    @patch('sys.stdout', new_callable=StringIO)
    def testTestableCellWithDependency(self, capturedStdout):
        expected = 100
        NotebookBuilder("notebook.ipynb", self.TEST_FILE_DIRECTORY) \
            .addMarkdownCell("# Markdown Cell") \
            .addCodeCell("imports", f"""
value = {expected}
            """, runnable=False) \
            .addMarkdownCell("# Markdown Cell") \
            .addCodeCell("testable", """""
print(value)
            """, deps=["imports"]) \
            .addMarkdownCell("# Markdown Cell") \
            .toFile()

        submission = IPythonSubmission()\
            .setSubmissionRoot(self.TEST_FILE_DIRECTORY)\
            .load()\
            .build()\
            .validate()

        cells = submission.getCells()

        self.assertIn("testable", cells.keys())
        self.assertEqual("imports", cells["testable"].metadata.deps[0])

        exec(submission.getExecutableSubmission())

        self.assertEqual(f"{expected}\n", capturedStdout.getvalue())

    def testTestableCellMissingDependency(self):
        expected = 100
        NotebookBuilder("notebook.ipynb", self.TEST_FILE_DIRECTORY) \
            .addMarkdownCell("# Markdown Cell") \
            .addCodeCell("imports", f"""
value = {expected}
            """, runnable=False) \
            .addMarkdownCell("# Markdown Cell") \
            .addCodeCell("testable", """
print(value)
            """, deps=["dne"]) \
            .addMarkdownCell("# Markdown Cell") \
            .toFile()

        with self.assertRaises(ValidationError) as ex:
            IPythonSubmission() \
                .setSubmissionRoot(self.TEST_FILE_DIRECTORY) \
                .load()

        msg = str(ex.exception)

        self.assertIn("missing dependency 'dne' for cell 'testable'", msg.lower())

    def testOnlyRunnableCellsCanBeActivated(self):
        NotebookBuilder("notebook.ipynb", self.TEST_FILE_DIRECTORY) \
            .addMarkdownCell("# Markdown Cell") \
            .addCodeCell("testableCell", self.VALID_TESTABLE_CELL) \
            .addMarkdownCell("# Markdown Cell") \
            .addCodeCell("validCell", self.VALID_CELL, runnable=False) \
            .addMarkdownCell("# Markdown Cell") \
            .toFile()

        submission = IPythonSubmission() \
            .setSubmissionRoot(self.TEST_FILE_DIRECTORY) \
            .load()\
            .build()\
            .validate()

        with self.assertRaises(RuntimeError) as ex:
            submission.setActiveCell("cell")

        msg = str(ex.exception)

        self.assertIn("invalid active cell", msg.lower())

    def testActiveCellMustBeSetWhenManyUnset(self):
        NotebookBuilder("notebook.ipynb", self.TEST_FILE_DIRECTORY) \
            .addMarkdownCell("# Markdown Cell") \
            .addCodeCell("imports", "", runnable=False) \
            .addMarkdownCell("# Markdown Cell") \
            .addCodeCell("testable", "", deps=["imports"]) \
            .addCodeCell("testableCell", self.VALID_TESTABLE_CELL) \
            .addMarkdownCell("# Markdown Cell") \
            .toFile()

        submission = IPythonSubmission() \
            .setSubmissionRoot(self.TEST_FILE_DIRECTORY) \
            .load() \
            .build() \
            .validate()

        with self.assertRaises(RuntimeError) as ex:
            submission.getExecutableSubmission()

        msg = str(ex.exception)

        self.assertIn("no active cell has been defined", msg.lower())

    @patch('sys.stdout', new_callable=StringIO)
    def testActiveCellMustBeSetWhenManySet(self, capturedStdout):
        expected = "imported!"
        NotebookBuilder("notebook.ipynb", self.TEST_FILE_DIRECTORY) \
            .addMarkdownCell("# Markdown Cell") \
            .addCodeCell("imports", f"""
value = '{expected}'
            """, runnable=False) \
            .addMarkdownCell("# Markdown Cell") \
            .addCodeCell("testable", """
print(value)
            """, deps=["imports"]) \
            .addCodeCell("testableCell", self.VALID_TESTABLE_CELL) \
            .addMarkdownCell("# Markdown Cell") \
            .toFile()

        submission = IPythonSubmission() \
            .setSubmissionRoot(self.TEST_FILE_DIRECTORY) \
            .load() \
            .build() \
            .validate()

        submission.setActiveCell("testable")

        exec(submission.getExecutableSubmission())

        self.assertEqual(f"{expected}\n", capturedStdout.getvalue())


    def testGenerateHTMLFailsWhenDisabled(self):
        NotebookBuilder("notebook2.ipynb", self.TEST_FILE_DIRECTORY) \
            .addMarkdownCell("# Markdown Cell") \
            .addCodeCell("testableCell", self.VALID_TESTABLE_CELL) \
            .addMarkdownCell("# Markdown Cell") \
            .toFile()

        submission = IPythonSubmission()\
            .setSubmissionRoot(self.TEST_FILE_DIRECTORY)\
            .load()

        with self.assertRaises(RuntimeError) as ex:
            submission.getNotebookHtml()

        msg = str(ex.exception)

        self.assertIn("notebook html requested, but not available", msg.lower())

    def testGenerateHTML(self):
        NotebookBuilder("notebook2.ipynb", self.TEST_FILE_DIRECTORY) \
            .addMarkdownCell("# Markdown Cell") \
            .addCodeCell("testableCell", self.VALID_TESTABLE_CELL) \
            .addMarkdownCell("# Markdown Cell") \
            .toFile()

        submission = IPythonSubmission() \
            .setSubmissionRoot(self.TEST_FILE_DIRECTORY) \
            .enableHtmlTransformation()\
            .load()

        html = submission.getNotebookHtml()

        # we are going to assume that the generation worked and nbconvert doesn't gaslight us
        self.assertIsNotNone(html)

    def testTransformMagicCommand(self):
        NotebookBuilder("notebook.ipynb", self.TEST_FILE_DIRECTORY)\
            .addCodeCell("testable",
            """
%matplotlib inline
!pip install yippee
            """)\
            .toFile()

        submission = IPythonSubmission() \
            .setSubmissionRoot(self.TEST_FILE_DIRECTORY) \
            .load()

        cell = submission.getCells()["testable"]

        self.assertNotIn("%", cell.code)
        self.assertNotIn("!", cell.code)


    def testTransformMatplotlibShow(self):
        NotebookBuilder("notebook.ipynb", self.TEST_FILE_DIRECTORY) \
            .addCodeCell("testable",
            """
TestableCell(id="testable", deps=[])
plt.show()
matplotlib.pyplot.show()
            """) \
            .toFile()

        submission = IPythonSubmission() \
            .setSubmissionRoot(self.TEST_FILE_DIRECTORY) \
            .load()

        cell = submission.getCells()["testable"]

        self.assertNotIn("show()", cell.code)
        self.assertIn("fig_1.png", cell.code)
        self.assertIn("fig_2.png", cell.code)

    def testVerifyIgnoresHiddenCacheAndSpacedFiles(self):
        os.makedirs(os.path.join(self.TEST_FILE_DIRECTORY, "dir", "__pycache__"))

        NotebookBuilder(".notebook.ipynb", self.TEST_FILE_DIRECTORY) \
            .addCodeCell("testableCell", self.VALID_TESTABLE_CELL) \
            .toFile()

        NotebookBuilder("a notebook.ipynb", self.TEST_FILE_DIRECTORY) \
            .addCodeCell("testableCell", self.VALID_TESTABLE_CELL) \
            .toFile()

        NotebookBuilder("a notebook.ipynb", self.TEST_FILE_DIRECTORY) \
            .addCodeCell("testableCell", self.VALID_TESTABLE_CELL) \
            .toFile()

        NotebookBuilder("notebook.ipynb", self.TEST_FILE_DIRECTORY) \
            .addCodeCell("testableCell", self.VALID_TESTABLE_CELL) \
            .toFile()

        submission = IPythonSubmission() \
            .setSubmissionRoot(self.TEST_FILE_DIRECTORY)\
            .load()

        cells = submission.getCells()

        self.assertEqual(1, len(cells))


    @patch("sys.stdout", new_callable=StringIO)
    def testInstallPackages(self, capturedStdout):
        NotebookBuilder("notebook.ipynb", self.TEST_FILE_DIRECTORY) \
            .addCodeCell("testable", """
import pip_install_test
            """) \
            .toFile()

        submission = IPythonSubmission() \
            .setSubmissionRoot(self.TEST_FILE_DIRECTORY) \
            .addPackages([
                {"name": "pip-install-test", "version": "0.5"},
                {"name": "minimal", "version": ""}
            ])\
            .load()\
            .build()\
            .validate()

        exec(submission.getExecutableSubmission())

        self.assertIn("good job", capturedStdout.getvalue().lower())

        submission.TEST_ONLY_removeRequirements()

    def testPackageDNE(self):
        NotebookBuilder("notebook.ipynb", self.TEST_FILE_DIRECTORY) \
            .addCodeCell("testable", """
import pip_install_test
            """) \
            .toFile()

        with self.assertRaises(ValidationError) as ex:
            IPythonSubmission() \
                .setSubmissionRoot(self.TEST_FILE_DIRECTORY) \
                .addPackages([
                {"name": "dne", "version": "0.5"},
                {"name": "dne2", "version": ""},
            ]) \
                .load() \
                .build()

        msg = str(ex.exception)

        self.assertIn("Unable to locate package, 'dne'", msg)
        self.assertIn("Unable to locate package, 'dne2'", msg)



from typing import Dict, List, Optional, Any

from nbformat import read, NotebookNode, write
from rich.console import Console
from rich.syntax import Syntax

from autograder_platform.cli import AutograderCLITool
from autograder_platform.config.Config import AutograderConfigurationBuilder, AutograderConfiguration
from language_binds.IPython.metadata import CellMetadata


class IPythonSubmissionManager(AutograderCLITool):
    def __init__(self):
        super().__init__(f"IPython Submission Manager v{AutograderCLITool.get_version()}")


    def configure_options(self):
        self.parser.add_argument("--edit", default=False, action="store_true",
                                 help="Run the tool in edit mode")
        self.parser.add_argument("--new-from-template", default=False, action="store_true",
                                 help="Run the tool in 'new-from-template' mode which allows injecting headers and stripping out solution code")
        self.parser.add_argument("--header-file", default="./header.md",
                                 help="The header that should be included in the outputted ipython file")
        self.parser.add_argument("--in-file", default="./submission.ipynb",
                                 help="ipython file that should be used.")
        self.parser.add_argument("--output", default="./bin/submission.ipynb",
                                 help="Where the outputted file should be")


    def set_config_arguments(self, configBuilder: AutograderConfigurationBuilder[AutograderConfiguration]):
        pass


    def print_cell_src(self, src: List[str]):
        joined = "".join(src)

        syntax = Syntax(joined, "python", line_numbers=True)
        console = Console()

        console.print(syntax)

    @staticmethod
    def edit_cell_metadata(cell_metadata: Dict[str, Dict[str, Any]]):
        if "autograder" in cell_metadata:
            metadata = CellMetadata(**cell_metadata["autograder"])
        else:
            metadata = CellMetadata()

        print(f"Current Metadata: {metadata}")

        print(f"Enter field name to edit, 'done' when you are done, or 'ignore' if no metadata should be written for this cell.")
        user_in = ""

        while user_in != "done":
            user_in = input("> ").lower()

            if user_in == "done":
                break

            if user_in == "ignore":
                return

            if user_in not in metadata.__dict__:
                print(f"{user_in} not in {metadata.__dict__.keys()}")
                continue
            if user_in == "deps":
                print("Enter the cell dependencies in order and 'done' once you are finished")
                deps = []
                dep = ""
                while dep.lower() != "done":
                    dep = input("deps > ")
                    if dep.lower() == "done":
                        break
                    deps.append(dep)
                to_set = deps

            else:
                to_set = input(f"{user_in} > ")

            if user_in == "runnable":
                to_set = bool(to_set)

            metadata.__dict__[user_in] = to_set

        cell_metadata["autograder"] = metadata.__dict__

    def edit(self, inputFile: str) -> NotebookNode:
        self.print_info_message(f"Editing '{inputFile}'")
        with open(inputFile, 'r', encoding="UTF-8") as r:
            notebook = read(r, as_version=4)

        for cell in notebook["cells"]:
            if cell["cell_type"] != "code":
                continue
            if "metadata" not in cell:
                cell["metadata"] = {}

            self.print_info_message("Editing metadata for cell: ")
            self.print_cell_src(cell["source"])
            self.edit_cell_metadata(cell["metadata"])

        return notebook


    def run(self) -> bool:
        self.configure_options()

        self.arguments = self.parser.parse_args()

        if self.arguments.version:
            self.print_info_message(f"Autograder Platform Version: {self.get_version()}")
            self.print_info_message(f"IPython Language Bind Version: TODO - a beta or something")
            return False

        notebook: Optional[NotebookNode] = None
        if self.arguments.edit:
            notebook = self.edit(self.arguments.in_file)
        elif self.arguments.new_from_template:
            self.print_error_message("NotImplemented", "Creating a new submission from a template has not yet been implemented.")

        if notebook is None:
            return False

        self.print_info_message(f"Writing '{self.arguments.output}'")
        with open(self.arguments.output, 'w') as w:
            write(notebook, w)

        return False


tool = IPythonSubmissionManager().run

if __name__ == "__main__":
    res = tool()

    exit(res)



import dataclasses
import os
import re
import subprocess
import sys
from types import CodeType
from typing import TypeVar, List, Dict, Iterable, Optional

import nbconvert
from nbformat import read, NotebookNode

from autograder_platform.StudentSubmission.AbstractStudentSubmission import AbstractStudentSubmission
from language_binds.IPython.metadata import CellMetadata
from language_binds.IPython.IPythonTransformers import MagicCommandTransformer, \
    MatplotLibFigTransformer
from language_binds.IPython.IPythonValidators import IPythonFileValidator, \
    TestableCellValidator, TestableCellDependencyValidator

from language_binds.Python.PythonValidators import PackageValidator

Builder = TypeVar("Builder", bound="IPythonSubmission")


def filterSearchResults(path: str) -> bool:
    if path[0] == ".":
        return False
    if "__pycache__" in path:
        return False
    if " " in path:
        return False

    return True


@dataclasses.dataclass()
class Cell:
    metadata: CellMetadata
    code: str


class IPythonSubmission(AbstractStudentSubmission[CodeType]):
    IPYTHON_FILE_REGEX: re.Pattern = re.compile(r"^(\w|-)+\.ipynb")

    def __init__(self):
        super().__init__()
        self._cells: Dict[str, Cell] = {}
        self._discoveredFiles: List[str] = []
        self._htmlTransformationEnabled: bool = False
        self._notebookHtml: Optional[str] = None
        self._extraPackages: Dict[str, str] = {}

        self._builtCells: Dict[str, CodeType] = {}

        self._activeCell: Optional[str] = None

        self.addValidator(PackageValidator())
        self.addValidator(IPythonFileValidator())
        self.addValidator(TestableCellValidator())
        self.addValidator(TestableCellDependencyValidator())
        self.addTransformer(MagicCommandTransformer())
        self.addTransformer(MatplotLibFigTransformer())

    def _discoverSubmittedFiles(self, directoryToSearch: str):
        pathsToVisit: Iterable[str] = filter(filterSearchResults, os.listdir(directoryToSearch))

        if not pathsToVisit:
            return

        for path in pathsToVisit:
            if os.path.isdir(os.path.join(directoryToSearch, path)):
                self._discoverSubmittedFiles(os.path.join(directoryToSearch, path))
                continue

            if self.IPYTHON_FILE_REGEX.match(path):
                self._discoveredFiles.append(os.path.join(directoryToSearch, path))

    def _exportHTML(self, notebook: NotebookNode):
        if self._htmlTransformationEnabled:
            exporter = nbconvert.HTMLExporter()
            (self._notebookHtml, _) = exporter.from_notebook_node(notebook)

    def _parseCells(self, notebook: NotebookNode):
        for cell in notebook.cells:
            if cell["cell_type"] != "code":
                continue

            if "metadata" not in cell:
                continue

            if "autograder" not in cell["metadata"]:
                continue

            try:
                metadata = CellMetadata(**cell['metadata']['autograder'])
            except Exception:
                continue

            if metadata is None:
                continue

            source = self.runTransformers(cell["source"])

            cellWithMetadata = Cell(metadata, source)

            self._cells[metadata.id] = cellWithMetadata

    def addPackages(self: Builder, packages: List[Dict[str, str]]) -> Builder:
        for package in packages:
            self._extraPackages.update({package['name']: package['version']})

        return self

    def _installRequirements(self) -> None:
        for package, version in self._extraPackages.items():
            try:
                subprocess.check_call([sys.executable, "-m", "pip", "install",
                                       f"{package}=={version}" if version else package],
                                      stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            except subprocess.CalledProcessError as _:  # pragma: no cover
                try:  # pragma: no cover
                    subprocess.check_call([sys.executable, "-m", "pip", "install",  # pragma: no cover
                                           f"{package}=={version}" if version else package, "--break-system-packages"],
                                          # pragma: no cover

                                          stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)  # pragma: no cover
                except subprocess.CalledProcessError as error:  # pragma: no cover
                    raise Exception(f"Failed to install '{package}'!")  # pragma: no cover

    def doLoad(self):
        self._discoverSubmittedFiles(self.getSubmissionRoot())
        self.runManualValidationHook(IPythonFileValidator)

        fileToLoad = self._discoveredFiles[0]

        with open(fileToLoad, 'r', encoding="UTF-8") as r:
            loadedNotebook: NotebookNode = read(r, as_version=4)

        self._exportHTML(loadedNotebook)

        self._parseCells(loadedNotebook)

    def _buildCells(self):
        for cell in self._cells.values():
            if not cell.metadata.runnable:
                continue

            sortedDeps = cell.metadata.deps

            # testable code is always last in the ordering
            sortedDeps.append(cell.metadata.id)

            combinedSrc = "\n".join([self._cells[id].code for id in sortedDeps])

            builtCell: CodeType = compile(combinedSrc, f"student_submission_{cell.metadata.id}", "exec")

            self._builtCells[cell.metadata.id] = builtCell

    def doBuild(self):
        self._installRequirements()
        self._buildCells()

    def TEST_ONLY_removeRequirements(self):
        for package in self._extraPackages.keys():
            subprocess.check_call([sys.executable, "-m", "pip", "uninstall",
                                   "-y", package],
                                  stdout=subprocess.DEVNULL)

    def setActiveCell(self, activeCell: str):
        if activeCell not in self._builtCells.keys():
            raise RuntimeError(
                f"Invalid active cell '{activeCell}'! Expected one of {', '.join(self._builtCells.keys())}")

        self._activeCell = activeCell

    def getExecutableSubmission(self) -> CodeType:
        if self._activeCell is None and len(self._builtCells) != 1:
            raise RuntimeError("No active cell has been defined!")

        if len(self._builtCells) == 1:
            return self._builtCells[list(self._builtCells.keys())[0]]

        return self._builtCells[self._activeCell]

    def enableHtmlTransformation(self: Builder, enabled: bool = True) -> Builder:
        self._htmlTransformationEnabled = enabled
        return self

    def getCells(self) -> Dict[str, Cell]:
        return self._cells

    def getExtraPackages(self) -> Dict[str, str]:
        return self._extraPackages

    def getDiscoveredFiles(self) -> List[str]:
        return self._discoveredFiles

    def getNotebookHtml(self) -> str:
        if not self._htmlTransformationEnabled or self._notebookHtml is None:
            raise RuntimeError("Notebook HTML requested, but not available!")
        return self._notebookHtml



from typing import Iterable


class MissingNotebookFile(Exception):
    def __init__(self) -> None:
        super().__init__(f"No `.ipynb` files were submitted! Expected exactly 1!")

class TooManyNotebooksError(Exception):
    def __init__(self, files: Iterable[str]) -> None:
        super().__init__(
            f"Expected one `.ipynb` file. Received: {', '.join(file for file in files)}\n"
            "Please delete extra `.ipynb` files"
        )

class MissingDependencyError(Exception):
    def __init__(self, cellId: str, dep: str, availableCells: Iterable[str]):
        super().__init__(
            f"Missing dependency '{dep}' for cell '{cellId}'!\n"
            f"Expected '{dep}' in {', '.join(cell for cell in availableCells)}, but was not present!"
        )

class NoTestableCellsError(Exception):
    def __init__(self):
        super().__init__(
            "Expected at least one cell to be testable!\n"
            "Currently, no cells are testable, please ensure at least one cell in notebook has 'TestableCell' metadata defined!"
        )



from typing import List

from autograder_platform.StudentSubmission.AbstractValidator import AbstractValidator
from autograder_platform.StudentSubmission.common import ValidationHook
from language_binds.IPython.common import MissingNotebookFile, TooManyNotebooksError, \
    MissingDependencyError, NoTestableCellsError

class TestableCellDependencyValidator(AbstractValidator):
    @staticmethod
    def getValidationHook() -> ValidationHook:
        return ValidationHook.POST_LOAD

    def __init__(self):
        super().__init__()
        self.cells = {}

    def setup(self, studentSubmission):
        self.cells = studentSubmission.getCells()

    def run(self):
        availableCells = self.cells.keys()

        for cell in self.cells.values():
            metadata = cell.metadata
            for dep in metadata.deps:
                if dep not in availableCells:
                    self.addError(MissingDependencyError(metadata.id, dep, availableCells))

class TestableCellValidator(AbstractValidator):
    @staticmethod
    def getValidationHook() -> ValidationHook:
        return ValidationHook.POST_LOAD

    def __init__(self):
        super().__init__()
        self.cells = {}

    def setup(self, studentSubmission):
        self.cells = studentSubmission.getCells()

    def run(self):
        numberRunnable = 0
        for cell in self.cells.values():
            metadata = cell.metadata

            numberRunnable += int(metadata.runnable)

        if numberRunnable == 0:
            self.addError(NoTestableCellsError())


class IPythonFileValidator(AbstractValidator):

    @staticmethod
    def getValidationHook() -> ValidationHook:
        return ValidationHook.MANUAL

    def __init__(self):
        super().__init__()
        self.files: List[str] = []

    def setup(self, studentSubmission):
        self.files = studentSubmission.getDiscoveredFiles()

    def run(self):
        if len(self.files) == 0:
            self.addError(MissingNotebookFile())
            return
        if len(self.files) > 1:
            self.addError(TooManyNotebooksError(self.files))



from autograder_platform.Registration.LanguageRegistration import LanguageRegistrationMetadata, register_language
from language_binds.Python import PythonConfigSchema
from language_binds.Python.PythonEnvironment import PythonEnvironment, configMapper

from language_binds.IPython.IPythonSubmission import IPythonSubmission
from language_binds.Python.PythonSubmissionProcess import RunnableStudentSubmission

__version__ = "1.0.0.RC-2"

metadata = LanguageRegistrationMetadata(
    name="ipython",
    version=__version__,
    on_submission_process_factory_registration= \
        lambda: (IPythonSubmission, RunnableStudentSubmission, PythonEnvironment, configMapper),
    on_config_registration=PythonConfigSchema,
)

register_language(metadata)

